@@ -3,6 +3,7 @@
 
 #include <QString>
 #include <QLocale>
+#include "common/qqmljssourcelocation_p.h"
 #include "qqmljsast_p.h"
 
 #include "qqmljsastvisitor_p.h"
@@ -1015,13 +1016,9 @@ BoundNames FormalParameterList::formals() const
                 // change the name of the earlier argument to enforce the lookup semantics from the spec
                 formals[duplicateIndex].id += QLatin1String("#") + QString::number(i);
             }
-            formals += {
-                    name,
-                    it->element->typeAnnotation,
-                    it->element->isInjectedSignalParameter
-                        ? BoundName::Injected
-                        : BoundName::Declared
-            };
+            formals += { name, it->element->firstSourceLocation(), it->element->typeAnnotation,
+                         it->element->isInjectedSignalParameter ? BoundName::Injected
+                                                                : BoundName::Declared };
         }
         ++i;
     }
@@ -1049,17 +1046,10 @@ void FormalParameterList::accept0(BaseVisitor *visitor)
     }
 }
 
-FormalParameterList *FormalParameterList::finish(QQmlJS::MemoryPool *pool)
+FormalParameterList *FormalParameterList::finish(QQmlJS::MemoryPool *)
 {
     FormalParameterList *front = next;
     next = nullptr;
-
-    int i = 0;
-    for (const FormalParameterList *it = this; it; it = it->next) {
-        if (it->element && it->element->bindingIdentifier.isEmpty())
-            it->element->bindingIdentifier = pool->newString(QLatin1String("arg#") + QString::number(i));
-        ++i;
-    }
     return front;
 }
 
@@ -1311,17 +1301,7 @@ void Type::accept0(BaseVisitor *visitor)
 {
     if (visitor->visit(this)) {
         accept(typeId, visitor);
-        accept(typeArguments, visitor);
-    }
-
-    visitor->endVisit(this);
-}
-
-void TypeArgumentList::accept0(BaseVisitor *visitor)
-{
-    if (visitor->visit(this)) {
-        for (TypeArgumentList *it = this; it; it = it->next)
-            accept(it->typeId, visitor);
+        accept(typeArgument, visitor);
     }
 
     visitor->endVisit(this);
@@ -1346,6 +1326,15 @@ void UiImport::accept0(BaseVisitor *visitor)
     visitor->endVisit(this);
 }
 
+void UiPragmaValueList::accept0(BaseVisitor *visitor)
+{
+    if (visitor->visit(this)) {
+    }
+
+    visitor->endVisit(this);
+}
+
+
 void UiPragma::accept0(BaseVisitor *visitor)
 {
     if (visitor->visit(this)) {
@@ -1405,6 +1394,11 @@ void TaggedTemplate::accept0(BaseVisitor *visitor)
     visitor->endVisit(this);
 }
 
+void InitializerExpression::accept0(BaseVisitor *visitor)
+{
+    expression->accept0(visitor);
+}
+
 void PatternElement::accept0(BaseVisitor *visitor)
 {
     if (visitor->visit(this)) {
@@ -1424,8 +1418,8 @@ void PatternElement::boundNames(BoundNames *names)
         else if (PatternPropertyList *p = propertyList())
             p->boundNames(names);
     } else {
-        names->append({bindingIdentifier.toString(), typeAnnotation,
-                       isInjectedSignalParameter ? BoundName::Injected : BoundName::Declared});
+        names->append({ bindingIdentifier.toString(), firstSourceLocation(), typeAnnotation,
+                        isInjectedSignalParameter ? BoundName::Injected : BoundName::Declared });
     }
 }
 
@@ -1563,17 +1557,11 @@ QString Type::toString() const
 
 void Type::toString(QString *out) const
 {
-    for (QQmlJS::AST::UiQualifiedId *it = typeId; it; it = it->next) {
-        out->append(it->name);
+    typeId->toString(out);
 
-        if (it->next)
-            out->append(QLatin1Char('.'));
-    }
-
-    if (typeArguments) {
+    if (typeArgument) {
         out->append(QLatin1Char('<'));
-        if (auto subType = static_cast<TypeArgumentList*>(typeArguments)->typeId)
-            subType->toString(out);
+        typeArgument->toString(out);
         out->append(QLatin1Char('>'));
     };
 }
