@@ -49,6 +49,8 @@
 QT_BEGIN_NAMESPACE
 
 Q_LOGGING_CATEGORY(lcRootProperties, "qt.qml.rootObjectProperties");
+Q_LOGGING_CATEGORY(lcQml, "qml");
+Q_LOGGING_CATEGORY(lcJs, "js");
 
 using namespace QV4;
 
@@ -85,6 +87,7 @@ the \l Qt::LeftButton and \l Qt::RightButton enumeration values as \c Qt.LeftBut
 
 
 \section1 Types
+\target globalqtobjecttypes
 
 The Qt object also contains helper functions for creating objects of specific
 data types. This is primarily useful when setting the properties of an item
@@ -107,8 +110,6 @@ creating objects of specific data types are also available for clients to use:
 \li \c matrix4x4 - use \l{Qt::matrix4x4()}{Qt.matrix4x4()}
 \endlist
 
-There are also string based constructors for these types. See \l{qtqml-typesystem-valuetypes.html}{QML Value Types} for more information.
-
 \section1 Date/Time Formatters
 
 The Qt object contains several functions for formatting QDateTime, QDate and QTime values.
@@ -188,7 +189,7 @@ The following functions are also on the Qt object.
 */
 
 /*!
-    \qmlproperty object Qt::application
+    \qmlproperty Application Qt::application
     \since 5.1
 
     The \c application object provides access to global application state
@@ -211,7 +212,7 @@ The following functions are also on the Qt object.
 */
 
 /*!
-    \qmlproperty object Qt::inputMethod
+    \qmlproperty InputMethod Qt::inputMethod
     \since 5.0
 
     It is the same as the \l InputMethod singleton.
@@ -267,13 +268,22 @@ the same object as is returned from the Qt.include() call.
 QtObject::QtObject(ExecutionEngine *engine)
     : m_engine(engine)
 {
-    QV4::Scope scope(engine);
-    QV4::ScopedString callLaterName(scope, engine->newIdentifier(QStringLiteral("callLater")));
+}
 
-    QV4::ScopedFunctionObject function(scope, QV4::FunctionObject::createBuiltinFunction(
-                engine, callLaterName, QQmlDelayedCallQueue::addUniquelyAndExecuteLater, 1));
+QtObject::Contexts QtObject::getContexts() const
+{
+    QQmlEngine *engine = qmlEngine();
+    if (!engine)
+        return {};
 
-    m_callLater = QJSValuePrivate::fromReturnedValue(function.asReturnedValue());
+    QQmlRefPointer<QQmlContextData> context = v4Engine()->callingQmlContext();
+    if (!context)
+        context = QQmlContextData::get(QQmlEnginePrivate::get(engine)->rootContext);
+
+    Q_ASSERT(context);
+    QQmlRefPointer<QQmlContextData> effectiveContext
+            = context->isPragmaLibraryContext() ? nullptr : context;
+    return {context, effectiveContext};
 }
 
 QtObject *QtObject::create(QQmlEngine *, QJSEngine *jsEngine)
@@ -464,11 +474,10 @@ QVariant QtObject::font(const QJSValue &fontSpecifier) const
     }
 
     {
-        QVariant v;
-        if (QQml_valueTypeProvider()->createValueType(
-                    QMetaType(QMetaType::QFont), fontSpecifier, v)) {
+        const QVariant v = QQmlValueTypeProvider::createValueType(
+                    fontSpecifier, QMetaType(QMetaType::QFont));
+        if (v.isValid())
             return v;
-        }
     }
 
     v4Engine()->throwError(QStringLiteral("Qt.font(): Invalid argument: "
@@ -496,15 +505,14 @@ void addParameters(QJSEngine *e, QJSValue &result, int i, T parameter, Others...
 }
 
 template<typename ...T>
-static QVariant createValueType(QJSEngine *e, QMetaType type, T... parameters)
+static QVariant constructFromJSValue(QJSEngine *e, QMetaType type, T... parameters)
 {
     if (!e)
         return QVariant();
     QJSValue params = e->newArray(sizeof...(parameters));
     addParameters(e, params, 0, parameters...);
-    QVariant variant;
-    QQml_valueTypeProvider()->createValueType(type, params, variant);
-    return variant;
+    const QVariant variant = QQmlValueTypeProvider::createValueType(params, type);
+    return variant.isValid() ? variant : QVariant(type);
 }
 
 /*!
@@ -514,7 +522,7 @@ static QVariant createValueType(QJSEngine *e, QMetaType type, T... parameters)
 */
 QVariant QtObject::vector2d(double x, double y) const
 {
-    return createValueType(jsEngine(), QMetaType(QMetaType::QVector2D), x, y);
+    return constructFromJSValue(jsEngine(), QMetaType(QMetaType::QVector2D), x, y);
 }
 
 /*!
@@ -524,7 +532,7 @@ QVariant QtObject::vector2d(double x, double y) const
 */
 QVariant QtObject::vector3d(double x, double y, double z) const
 {
-    return createValueType(jsEngine(), QMetaType(QMetaType::QVector3D), x, y, z);
+    return constructFromJSValue(jsEngine(), QMetaType(QMetaType::QVector3D), x, y, z);
 }
 
 /*!
@@ -534,7 +542,7 @@ QVariant QtObject::vector3d(double x, double y, double z) const
 */
 QVariant QtObject::vector4d(double x, double y, double z, double w) const
 {
-    return createValueType(jsEngine(), QMetaType(QMetaType::QVector4D), x, y, z, w);
+    return constructFromJSValue(jsEngine(), QMetaType(QMetaType::QVector4D), x, y, z, w);
 }
 
 /*!
@@ -544,7 +552,48 @@ QVariant QtObject::vector4d(double x, double y, double z, double w) const
 */
 QVariant QtObject::quaternion(double scalar, double x, double y, double z) const
 {
-    return createValueType(jsEngine(), QMetaType(QMetaType::QQuaternion), scalar, x, y, z);
+    return constructFromJSValue(jsEngine(), QMetaType(QMetaType::QQuaternion), scalar, x, y, z);
+}
+
+/*!
+    \qmlmethod matrix4x4 Qt::matrix4x4()
+
+    Returns an identity matrix4x4.
+ */
+QVariant QtObject::matrix4x4() const
+{
+    const QMetaType metaType(QMetaType::QMatrix4x4);
+    const QVariant variant = QQmlValueTypeProvider::createValueType(QJSValue(), metaType);
+    return variant.isValid() ? variant : QVariant(metaType);
+}
+
+/*!
+    \qmlmethod matrix4x4 Qt::matrix4x4(var values)
+
+    Returns a matrix4x4 with the specified \a values. \a values is expected to
+    be a JavaScript array with 16 entries.
+
+    The array indices correspond to positions in the matrix as follows:
+
+    \table
+    \row \li 0  \li 1  \li 2  \li 3
+    \row \li 4  \li 5  \li 6  \li 7
+    \row \li 8  \li 9  \li 10 \li 11
+    \row \li 12 \li 13 \li 14 \li 15
+    \endtable
+*/
+QVariant QtObject::matrix4x4(const QJSValue &value) const
+{
+    if (value.isObject()) {
+        QVariant v = QQmlValueTypeProvider::createValueType(
+                    value, QMetaType(QMetaType::QMatrix4x4));
+        if (v.isValid())
+            return v;
+    }
+
+    v4Engine()->throwError(QStringLiteral("Qt.matrix4x4(): Invalid argument: "
+                                          "not a valid matrix4x4 values array"));
+    return QVariant();
 }
 
 /*!
@@ -560,41 +609,13 @@ QVariant QtObject::quaternion(double scalar, double x, double y, double z) const
     \row \li \a m31 \li \a m32 \li \a m33 \li \a m34
     \row \li \a m41 \li \a m42 \li \a m43 \li \a m44
     \endtable
-
-    Alternatively, the function may be called with a single argument
-    where that argument is a JavaScript array which contains the sixteen
-    matrix values.
-
-    Finally, the function may be called with no arguments and the resulting
-    matrix will be the identity matrix.
 */
-QVariant QtObject::matrix4x4() const
-{
-    QVariant variant;
-    QQml_valueTypeProvider()->createValueType(
-                QMetaType(QMetaType::QMatrix4x4), QJSValue(), variant);
-    return variant;
-}
-
-QVariant QtObject::matrix4x4(const QJSValue &value) const
-{
-    if (value.isObject()) {
-        QVariant v;
-        if (QQml_valueTypeProvider()->createValueType(QMetaType(QMetaType::QMatrix4x4), value, v))
-            return v;
-    }
-
-    v4Engine()->throwError(QStringLiteral("Qt.matrix4x4(): Invalid argument: "
-                                          "not a valid matrix4x4 values array"));
-    return QVariant();
-}
-
 QVariant QtObject::matrix4x4(double m11, double m12, double m13, double m14,
                              double m21, double m22, double m23, double m24,
                              double m31, double m32, double m33, double m34,
                              double m41, double m42, double m43, double m44) const
 {
-    return createValueType(jsEngine(), QMetaType(QMetaType::QMatrix4x4),
+    return constructFromJSValue(jsEngine(), QMetaType(QMetaType::QMatrix4x4),
                            m11, m12, m13, m14, m21, m22, m23, m24,
                            m31, m32, m33, m34, m41, m42, m43, m44);
 }
@@ -735,6 +756,11 @@ QString formatDateTimeObjectUsingDateFormat(T formatThis, Qt::DateFormat format)
 }
 }
 
+static QTime dateTimeToTime(const QDateTime &dateTime)
+{
+    return dateTime.toLocalTime().time();
+}
+
 /*!
 \qmlmethod string Qt::formatDate(datetime date, variant format, variant localeFormatOption)
 
@@ -751,22 +777,93 @@ default locale.
 
 \sa Locale
 */
-QString QtObject::formatDate(const QDate &date, const QString &format) const
+static std::optional<QDate> dateFromString(const QString &string, QV4::ExecutionEngine *engine)
+{
+    {
+        const QDate date = QDate::fromString(string, Qt::ISODate);
+        if (date.isValid())
+            return date;
+    }
+
+    {
+        // For historical reasons, the string argument is parsed as datetime, not as only date
+        const QDateTime dateTime = QDateTime::fromString(string, Qt::ISODate);
+        if (dateTime.isValid()) {
+            qCWarning(lcRootProperties())
+                    << string << "is a date/time string being passed to formatDate()."
+                    << "You should only pass date strings to formatDate().";
+            return dateTime.date();
+        }
+    }
+
+    {
+        // Since we can coerce QDate to QString, allow the resulting string format here.
+        const QDateTime dateTime = DateObject::stringToDateTime(string, engine);
+        if (dateTime.isValid())
+            return DateObject::dateTimeToDate(dateTime);
+    }
+
+    engine->throwError(QStringLiteral("Invalid argument passed to formatDate(): %1").arg(string));
+    return std::nullopt;
+}
+
+QString QtObject::formatDate(QDate date, const QString &format) const
 {
     return date.toString(format);
 }
 
-QString QtObject::formatDate(const QDate &date, Qt::DateFormat format) const
+QString QtObject::formatDate(QDate date, Qt::DateFormat format) const
 {
     return formatDateTimeObjectUsingDateFormat(date, format);
 }
 
+QString QtObject::formatDate(const QDateTime &dateTime, const QString &format) const
+{
+    return DateObject::dateTimeToDate(dateTime).toString(format);
+}
+
+QString QtObject::formatDate(const QString &string, const QString &format) const
+{
+    if (const auto qDate = dateFromString(string, v4Engine()))
+        return formatDate(qDate.value(), format);
+
+    return QString();
+}
+
+QString QtObject::formatDate(const QDateTime &dateTime, Qt::DateFormat format) const
+{
+    return formatDateTimeObjectUsingDateFormat(DateObject::dateTimeToDate(dateTime), format);
+}
+
+QString QtObject::formatDate(const QString &string, Qt::DateFormat format) const
+{
+    if (const auto qDate = dateFromString(string, v4Engine()))
+        return formatDate(qDate.value(), format);
+
+    return QString();
+}
+
 #if QT_CONFIG(qml_locale)
-QString QtObject::formatDate(const QDate &date, const QLocale &locale,
+QString QtObject::formatDate(QDate date, const QLocale &locale,
                              QLocale::FormatType formatType) const
 {
     return locale.toString(date, formatType);
 }
+
+QString QtObject::formatDate(const QDateTime &dateTime, const QLocale &locale,
+                             QLocale::FormatType formatType) const
+{
+    return locale.toString(DateObject::dateTimeToDate(dateTime), formatType);
+}
+
+QString QtObject::formatDate(const QString &string, const QLocale &locale,
+                             QLocale::FormatType formatType) const
+{
+    if (const auto qDate = dateFromString(string, v4Engine()))
+        return locale.toString(qDate.value(), formatType);
+
+    return QString();
+}
 #endif
 
 /*!
@@ -804,15 +901,27 @@ static std::optional<QTime> timeFromString(const QString &string, QV4::Execution
         }
     }
 
+    {
+        // Since we can coerce QTime to QString, allow the resulting string format here.
+        const QDateTime dateTime = DateObject::stringToDateTime(string, engine);
+        if (dateTime.isValid())
+            return dateTimeToTime(dateTime);
+    }
+
     engine->throwError(QStringLiteral("Invalid argument passed to formatTime(): %1").arg(string));
     return std::nullopt;
 }
 
-QString QtObject::formatTime(const QTime &time, const QString &format) const
+QString QtObject::formatTime(QTime time, const QString &format) const
 {
     return time.toString(format);
 }
 
+QString QtObject::formatTime(const QDateTime &dateTime, const QString &format) const
+{
+    return dateTimeToTime(dateTime).toString(format);
+}
+
 QString QtObject::formatTime(const QString &time, const QString &format) const
 {
 
@@ -822,11 +931,16 @@ QString QtObject::formatTime(const QString &time, const QString &format) const
     return QString();
 }
 
-QString QtObject::formatTime(const QTime &time, Qt::DateFormat format) const
+QString QtObject::formatTime(QTime time, Qt::DateFormat format) const
 {
     return formatDateTimeObjectUsingDateFormat(time, format);
 }
 
+QString QtObject::formatTime(const QDateTime &dateTime, Qt::DateFormat format) const
+{
+    return formatDateTimeObjectUsingDateFormat(dateTimeToTime(dateTime), format);
+}
+
 QString QtObject::formatTime(const QString &time, Qt::DateFormat format) const
 {
     if (auto qTime = timeFromString(time, v4Engine()))
@@ -836,12 +950,18 @@ QString QtObject::formatTime(const QString &time, Qt::DateFormat format) const
 }
 
 #if QT_CONFIG(qml_locale)
-QString QtObject::formatTime(const QTime &time, const QLocale &locale,
+QString QtObject::formatTime(QTime time, const QLocale &locale,
                              QLocale::FormatType formatType) const
 {
     return locale.toString(time, formatType);
 }
 
+QString QtObject::formatTime(const QDateTime &dateTime, const QLocale &locale,
+                             QLocale::FormatType formatType) const
+{
+    return locale.toString(dateTimeToTime(dateTime), formatType);
+}
+
 QString QtObject::formatTime(const QString &time, const QLocale &locale,
                              QLocale::FormatType formatType) const
 {
@@ -950,22 +1070,69 @@ with the \a format values below to produce the following results:
 
     \sa Locale
 */
+static std::optional<QDateTime> dateTimeFromString(const QString &string, QV4::ExecutionEngine *engine)
+{
+    {
+        const QDateTime dateTime = QDateTime::fromString(string, Qt::ISODate);
+        if (dateTime.isValid())
+            return dateTime;
+    }
+
+    {
+        // Since we can coerce QDateTime to QString, allow the resulting string format here.
+        const QDateTime dateTime = DateObject::stringToDateTime(string, engine);
+        if (dateTime.isValid())
+            return dateTime;
+    }
+
+    engine->throwError(QStringLiteral("Invalid argument passed to formatDateTime(): %1").arg(string));
+    return std::nullopt;
+}
+
 QString QtObject::formatDateTime(const QDateTime &dateTime, const QString &format) const
 {
     return dateTime.toString(format);
 }
 
+QString QtObject::formatDateTime(const QString &string, const QString &format) const
+{
+
+    if (const auto qDateTime = dateTimeFromString(string, v4Engine()))
+        return formatDateTime(qDateTime.value(), format);
+
+    return QString();
+}
+
 QString QtObject::formatDateTime(const QDateTime &dateTime, Qt::DateFormat format) const
 {
     return formatDateTimeObjectUsingDateFormat(dateTime, format);
 }
 
+QString QtObject::formatDateTime(const QString &string, Qt::DateFormat format) const
+{
+
+    if (const auto qDateTime = dateTimeFromString(string, v4Engine()))
+        return formatDateTime(qDateTime.value(), format);
+
+    return QString();
+}
+
 #if QT_CONFIG(qml_locale)
 QString QtObject::formatDateTime(const QDateTime &dateTime, const QLocale &locale,
                                  QLocale::FormatType formatType) const
 {
     return locale.toString(dateTime, formatType);
 }
+
+QString QtObject::formatDateTime(const QString &string, const QLocale &locale,
+                                 QLocale::FormatType formatType) const
+{
+
+    if (const auto qDateTime = dateTimeFromString(string, v4Engine()))
+        return formatDateTime(qDateTime.value(), locale, formatType);
+
+    return QString();
+}
 #endif
 
 /*!
@@ -1132,10 +1299,15 @@ Each object in this array has the members \c lineNumber, \c columnNumber, \c fil
 For example, if the above snippet had misspelled color as 'colro' then the array would contain an object like the following:
 { "lineNumber" : 1, "columnNumber" : 32, "fileName" : "dynamicSnippet1", "message" : "Cannot assign to non-existent property \"colro\""}.
 
-Note that this function returns immediately, and therefore may not work if
+\note This function returns immediately, and therefore may not work if
 the \a qml string loads new components (that is, external QML files that have not yet been loaded).
 If this is the case, consider using \l{QtQml::Qt::createComponent()}{Qt.createComponent()} instead.
 
+\warning This function is extremely slow since it has to compile the passed QML string every time
+it is invoked. Furthermore, it's very easy to produce invalid QML when programmatically constructing
+QML code. It's much better to keep your QML components as separate files and add properties and
+methods to customize their behavior than to produce new components by string manipulation.
+
 See \l {Dynamic QML Object Creation from JavaScript} for more information on using this function.
 */
 QObject *QtObject::createQmlObject(const QString &qml, QObject *parent, const QUrl &url) const
@@ -1158,7 +1330,7 @@ QObject *QtObject::createQmlObject(const QString &qml, QObject *parent, const QU
             QV4::ScopedObject qmlerror(scope);
             QV4::ScopedString s(scope);
             QV4::ScopedValue v(scope);
-            for (int ii = 0; ii < errors.count(); ++ii) {
+            for (int ii = 0; ii < errors.size(); ++ii) {
                 const QQmlError &error = errors.at(ii);
                 errorstr += QLatin1String("\n    ") + error.toString();
                 qmlerror = v4->newObject();
@@ -1234,7 +1406,7 @@ QObject *QtObject::createQmlObject(const QString &qml, QObject *parent, const QU
         obj->setParent(parent);
 
         QList<QQmlPrivate::AutoParentFunction> functions = QQmlMetaType::parentFunctions();
-        for (int ii = 0; ii < functions.count(); ++ii) {
+        for (int ii = 0; ii < functions.size(); ++ii) {
             if (QQmlPrivate::Parented == functions.at(ii)(obj, parent))
                 break;
         }
@@ -1295,6 +1467,32 @@ See \l {Dynamic QML Object Creation from JavaScript} for more information on usi
 To create a QML object from an arbitrary string of QML (instead of a file),
 use \l{QtQml::Qt::createQmlObject()}{Qt.createQmlObject()}.
 */
+
+/*!
+\qmlmethod Component Qt::createComponent(string moduleUri, string typeName, enumeration mode, QtObject parent)
+\overload
+Returns a \l Component object created for the type specified by \a moduleUri and \a typeName.
+\qml
+import QtQuick
+QtObject {
+    id: root
+    property Component myComponent: Qt.createComponent("QtQuick", "Rectangle", Component.Asynchronous, root)
+}
+\endqml
+This overload mostly behaves as the \c url based version, but can be used
+to instantiate types which do not have an URL (e.g. C++ types registered
+via \l {QML_ELEMENT}).
+\note In some cases, passing \c Component.Asynchronous won't have any
+effect:
+\list
+\li The type is implemented in C++
+\li The type is an inline component.
+\endlist
+If the optional \a parent parameter is given, it should refer to the object
+that will become the parent for the created \l Component object. If no mode
+was passed, this can be the second argument.
+*/
+
 QQmlComponent *QtObject::createComponent(const QUrl &url, QObject *parent) const
 {
     return createComponent(url, QQmlComponent::PreferSynchronous, parent);
@@ -1315,13 +1513,9 @@ QQmlComponent *QtObject::createComponent(const QUrl &url, QQmlComponent::Compila
     if (!engine)
         return nullptr;
 
-    QQmlRefPointer<QQmlContextData> context = v4Engine()->callingQmlContext();
+    auto [context, effectiveContext] = getContexts();
     if (!context)
-        context = QQmlContextData::get(QQmlEnginePrivate::get(engine)->rootContext);
-
-    Q_ASSERT(context);
-    QQmlRefPointer<QQmlContextData> effectiveContext
-            = context->isPragmaLibraryContext() ? nullptr : context;
+        return nullptr;
 
     QQmlComponent *c = new QQmlComponent(engine, context->resolvedUrl(url), mode, parent);
     QQmlComponentPrivate::get(c)->creationContext = effectiveContext;
@@ -1330,6 +1524,41 @@ QQmlComponent *QtObject::createComponent(const QUrl &url, QQmlComponent::Compila
     return c;
 }
 
+QQmlComponent *QtObject::createComponent(const QString &moduleUri, const QString &typeName,
+                                         QObject *parent) const
+{
+    return createComponent(moduleUri, typeName, QQmlComponent::PreferSynchronous, parent);
+}
+
+QQmlComponent *QtObject::createComponent(const QString &moduleUri, const QString &typeName, QQmlComponent::CompilationMode mode, QObject *parent) const
+{
+    if (mode != QQmlComponent::Asynchronous && mode != QQmlComponent::PreferSynchronous) {
+        v4Engine()->throwError(QStringLiteral("Invalid compilation mode %1").arg(mode));
+        return nullptr;
+    }
+
+    QQmlEngine *engine = qmlEngine();
+    if (!engine)
+        return nullptr;
+
+    if (moduleUri.isEmpty() || typeName.isEmpty())
+        return nullptr;
+
+    auto [context, effectiveContext] = getContexts();
+    if (!context)
+        return nullptr;
+
+    QQmlComponent *c = new QQmlComponent(engine, moduleUri, typeName, mode, parent);
+    if (c->isError() && !parent && moduleUri.endsWith(u".qml")) {
+        v4Engine()->throwTypeError(
+                    QStringLiteral("Invalid arguments; did you swap mode and parent"));
+    }
+    QQmlComponentPrivate::get(c)->creationContext = effectiveContext;
+    QQmlData::get(c, true)->explicitIndestructibleSet = false;
+    QQmlData::get(c)->indestructible = false;
+    return c;
+}
+
 #if QT_CONFIG(translation)
 QString QtObject::uiLanguage() const
 {
@@ -1459,6 +1688,11 @@ QJSValue QtObject::binding(const QJSValue &function) const
                 Encode(e->memoryManager->allocate<QQmlBindingFunction>(f)));
 }
 
+void QtObject::callLater(QQmlV4Function *args)
+{
+    m_engine->delayedCallQueue()->addUniquelyAndExecuteLater(m_engine, args);
+}
+
 
 QQmlPlatform *QtObject::platform()
 {
@@ -1486,11 +1720,6 @@ QObject *QtObject::styleHints() const
     return QQml_guiProvider()->styleHints();
 }
 
-QJSValue QtObject::callLater() const
-{
-    return m_callLater;
-}
-
 void QV4::Heap::ConsoleObject::init()
 {
     Object::init();
@@ -1524,21 +1753,22 @@ enum ConsoleLogTypes {
 static QString jsStack(QV4::ExecutionEngine *engine) {
     QString stack;
 
-    QVector<QV4::StackFrame> stackTrace = engine->stackTrace(10);
-
-    for (int i = 0; i < stackTrace.count(); i++) {
-        const QV4::StackFrame &frame = stackTrace.at(i);
-
+    int i = 0;
+    for (CppStackFrame *f = engine->currentStackFrame; f && i < 10; f = f->parentFrame(), ++i) {
         QString stackFrame;
-        if (frame.column >= 0)
-            stackFrame = QStringLiteral("%1 (%2:%3:%4)").arg(frame.function,
-                                                             frame.source,
-                                                             QString::number(frame.line),
-                                                             QString::number(frame.column));
-        else
-            stackFrame = QStringLiteral("%1 (%2:%3)").arg(frame.function,
-                                                             frame.source,
-                                                             QString::number(frame.line));
+
+        if (f->isJSTypesFrame() && static_cast<JSTypesStackFrame *>(f)->isTailCalling()) {
+            stackFrame = QStringLiteral("[elided tail calls]");
+        } else {
+            const int line = f->lineNumber();
+            if (line != f->missingLineNumber()) {
+                stackFrame = QStringLiteral("%1 (%2:%3)").arg(
+                        f->function(), f->source(), QString::number(qAbs(line)));
+            } else {
+                stackFrame = QStringLiteral("%1 (%2)").arg(
+                        f->function(), f->source());
+            }
+        }
 
         if (i)
             stack += QLatin1Char('\n');
@@ -1575,7 +1805,7 @@ static QString serializeArray(Object *array, ExecutionEngine *v4, QSet<QV4::Heap
 static ReturnedValue writeToConsole(const FunctionObject *b, const Value *argv, int argc,
                                     ConsoleLogTypes logType, bool printStack = false)
 {
-    QLoggingCategory *loggingCategory = nullptr;
+    const QLoggingCategory *loggingCategory = nullptr;
     QString result;
     QV4::Scope scope(b);
     QV4::ExecutionEngine *v4 = scope.engine;
@@ -1608,11 +1838,8 @@ static ReturnedValue writeToConsole(const FunctionObject *b, const Value *argv,
     if (printStack)
         result += QLatin1Char('\n') + jsStack(v4);
 
-    static QLoggingCategory qmlLoggingCategory("qml");
-    static QLoggingCategory jsLoggingCategory("js");
-
     if (!loggingCategory)
-        loggingCategory = v4->qmlEngine() ? &qmlLoggingCategory : &jsLoggingCategory;
+        loggingCategory = v4->qmlEngine() ? &lcQml() : &lcJs();
     QV4::CppStackFrame *frame = v4->currentStackFrame;
     const QByteArray baSource = frame ? frame->source().toUtf8() : QByteArray();
     const QByteArray baFunction = frame ? frame->function().toUtf8() : QByteArray();
@@ -1767,7 +1994,7 @@ ReturnedValue ConsoleObject::method_trace(const FunctionObject *b, const Value *
     QV4::CppStackFrame *frame = v4->currentStackFrame;
     QMessageLogger(frame->source().toUtf8().constData(), frame->lineNumber(),
                    frame->function().toUtf8().constData())
-        .debug("%s", qPrintable(stack));
+        .debug(v4->qmlEngine() ? lcQml : lcJs, "%s", qPrintable(stack));
 
     return QV4::Encode::undefined();
 }
@@ -1869,7 +2096,7 @@ void QV4::GlobalExtensions::init(Object *globalObject, QJSEngine::Extensions ext
     Example:
     \snippet qml/qsTranslate.qml 0
 
-    \sa {Internationalization and Localization with Qt Quick}
+    \sa {Internationalization with Qt}
 */
 ReturnedValue GlobalExtensions::method_qsTranslate(const FunctionObject *b, const Value *, const Value *argv, int argc)
 {
@@ -1911,7 +2138,7 @@ ReturnedValue GlobalExtensions::method_qsTranslate(const FunctionObject *b, cons
 }
 
 /*!
-    \qmlmethod string Qt::qsTranslateNoOp(string context, string sourceText, string disambiguation)
+    \qmlmethod string Qt::QT_TRANSLATE_NOOP(string context, string sourceText, string disambiguation)
 
     Marks \a sourceText for dynamic translation in the given \a context; i.e, the stored \a sourceText
     will not be altered.
@@ -1930,7 +2157,7 @@ ReturnedValue GlobalExtensions::method_qsTranslate(const FunctionObject *b, cons
     Example:
     \snippet qml/qtTranslateNoOp.qml 0
 
-    \sa {Internationalization and Localization with Qt Quick}
+    \sa {Internationalization with Qt}
 */
 ReturnedValue GlobalExtensions::method_qsTranslateNoOp(const FunctionObject *b, const Value *, const Value *argv, int argc)
 {
@@ -1941,6 +2168,46 @@ ReturnedValue GlobalExtensions::method_qsTranslateNoOp(const FunctionObject *b,
         return argv[1].asReturnedValue();
 }
 
+QString GlobalExtensions::currentTranslationContext(ExecutionEngine *engine)
+{
+    QString context;
+    CppStackFrame *frame = engine->currentStackFrame;
+
+    // The first non-empty source URL in the call stack determines the translation context.
+    while (frame && context.isEmpty()) {
+        if (ExecutableCompilationUnit *unit = frame->v4Function->executableCompilationUnit()) {
+            auto translationContextIndex = unit->data->translationContextIndex();
+            if (translationContextIndex)
+                context = unit->stringAt(*translationContextIndex);
+            if (!context.isEmpty())
+                break;
+            QString fileName = unit->fileName();
+            QUrl url(unit->fileName());
+            if (url.isValid() && url.isRelative()) {
+                context = url.fileName();
+            } else {
+                context = QQmlFile::urlToLocalFileOrQrc(fileName);
+                if (context.isEmpty() && fileName.startsWith(QLatin1String(":/")))
+                    context = fileName;
+            }
+            context = QFileInfo(context).completeBaseName();
+        }
+        frame = frame->parentFrame();
+    }
+
+    if (context.isEmpty()) {
+        if (QQmlRefPointer<QQmlContextData> ctxt = engine->callingQmlContext()) {
+            QString path = ctxt->urlString();
+            int lastSlash = path.lastIndexOf(QLatin1Char('/'));
+            int lastDot = path.lastIndexOf(QLatin1Char('.'));
+            int length = lastDot - (lastSlash + 1);
+            context = (lastSlash > -1) ? path.mid(lastSlash + 1, (length > -1) ? length : -1) : QString();
+        }
+    }
+
+    return context;
+}
+
 /*!
     \qmlmethod string Qt::qsTr(string sourceText, string disambiguation, int n)
 
@@ -1956,7 +2223,7 @@ ReturnedValue GlobalExtensions::method_qsTranslateNoOp(const FunctionObject *b,
     Example:
     \snippet qml/qsTr.qml 0
 
-    \sa {Internationalization and Localization with Qt Quick}
+    \sa {Internationalization with Qt}
 */
 ReturnedValue GlobalExtensions::method_qsTr(const FunctionObject *b, const Value *, const Value *argv, int argc)
 {
@@ -1970,43 +2237,10 @@ ReturnedValue GlobalExtensions::method_qsTr(const FunctionObject *b, const Value
     if ((argc > 2) && !argv[2].isNumber())
         THROW_GENERIC_ERROR("qsTr(): third argument (n) must be a number");
 
-    QString context;
-    CppStackFrame *frame = scope.engine->currentStackFrame;
-    // The first non-empty source URL in the call stack determines the translation context.
-    while (frame && context.isEmpty()) {
-        if (CompiledData::CompilationUnitBase *baseUnit = frame->v4Function->compilationUnit) {
-            const auto *unit = static_cast<const CompiledData::CompilationUnit *>(baseUnit);
-            QString fileName = unit->fileName();
-            QUrl url(unit->fileName());
-            if (url.isValid() && url.isRelative()) {
-                context = url.fileName();
-            } else {
-                context = QQmlFile::urlToLocalFileOrQrc(fileName);
-                if (context.isEmpty() && fileName.startsWith(QLatin1String(":/")))
-                    context = fileName;
-            }
-            context = QFileInfo(context).baseName();
-        }
-        frame = frame->parentFrame();
-    }
-
-    if (context.isEmpty()) {
-        if (QQmlRefPointer<QQmlContextData> ctxt = scope.engine->callingQmlContext()) {
-            QString path = ctxt->urlString();
-            int lastSlash = path.lastIndexOf(QLatin1Char('/'));
-            int lastDot = path.lastIndexOf(QLatin1Char('.'));
-            int length = lastDot - (lastSlash + 1);
-            context = (lastSlash > -1) ? path.mid(lastSlash + 1, (length > -1) ? length : -1) : QString();
-        }
-    }
-
-    QString text = argv[0].toQStringNoThrow();
-    QString comment;
-    if (argc > 1)
-        comment = argv[1].toQStringNoThrow();
-    int n = -1;
-    if (argc > 2)
-        n = argv[2].toInt32();
+    const QString context = currentTranslationContext(scope.engine);
+    const QString text = argv[0].toQStringNoThrow();
+    const QString comment = argc > 1 ? argv[1].toQStringNoThrow() : QString();
+    const int n = argc > 2 ? argv[2].toInt32() : -1;
 
     if (QQmlEnginePrivate *ep = (scope.engine->qmlEngine() ? QQmlEnginePrivate::get(scope.engine->qmlEngine()) : nullptr))
         if (ep->propertyCapture)
@@ -2019,7 +2253,7 @@ ReturnedValue GlobalExtensions::method_qsTr(const FunctionObject *b, const Value
 }
 
 /*!
-    \qmlmethod string Qt::qsTrNoOp(string sourceText, string disambiguation)
+    \qmlmethod string Qt::QT_TR_NOOP(string sourceText, string disambiguation)
 
     Marks \a sourceText for dynamic translation; i.e, the stored \a sourceText
     will not be altered.
@@ -2038,7 +2272,7 @@ ReturnedValue GlobalExtensions::method_qsTr(const FunctionObject *b, const Value
     Example:
     \snippet qml/qtTrNoOp.qml 0
 
-    \sa {Internationalization and Localization with Qt Quick}
+    \sa {Internationalization with Qt}
 */
 ReturnedValue GlobalExtensions::method_qsTrNoOp(const FunctionObject *, const Value *, const Value *argv, int argc)
 {
@@ -2076,7 +2310,7 @@ ReturnedValue GlobalExtensions::method_qsTrNoOp(const FunctionObject *, const Va
     Creating binary translation (QM) files suitable for use with this function requires passing
     the \c -idbased option to the \c lrelease tool.
 
-    \sa QT_TRID_NOOP(), {Internationalization and Localization with Qt Quick}
+    \sa QT_TRID_NOOP(), {Internationalization with Qt}
 */
 ReturnedValue GlobalExtensions::method_qsTrId(const FunctionObject *b, const Value *, const Value *argv, int argc)
 {
@@ -2100,7 +2334,7 @@ ReturnedValue GlobalExtensions::method_qsTrId(const FunctionObject *b, const Val
 }
 
 /*!
-    \qmlmethod string Qt::qsTrIdNoOp(string id)
+    \qmlmethod string Qt::QT_TRID_NOOP(string id)
 
     Marks \a id for dynamic translation.
 
@@ -2113,7 +2347,7 @@ ReturnedValue GlobalExtensions::method_qsTrId(const FunctionObject *b, const Val
     Example:
     \snippet qml/qtTrIdNoOp.qml 0
 
-    \sa qsTrId(), {Internationalization and Localization with Qt Quick}
+    \sa qsTrId(), {Internationalization with Qt}
 */
 ReturnedValue GlobalExtensions::method_qsTrIdNoOp(const FunctionObject *, const Value *, const Value *argv, int argc)
 {
@@ -2173,11 +2407,6 @@ be passed on to the function invoked. Note that if redundant calls
 are eliminated, then only the last set of arguments will be passed to the
 function.
 */
-ReturnedValue QtObject::method_callLater(const FunctionObject *b, const Value *thisObject,
-                                         const Value *argv, int argc)
-{
-    return b->engine()->delayedCallQueue()->addUniquelyAndExecuteLater(b, thisObject, argv, argc);
-}
 
 QT_END_NAMESPACE
 
